/*

bftpd Copyright (C) 1999-2000 Max-Wilhelm Bruker

This program is is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License, version 2 of the
License as published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

*/

#include <config.h>
#include <stdio.h>
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_ASM_SOCKET_H
#include <asm/socket.h>
#endif
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#ifdef HAVE_WAIT_H
# include <wait.h>
#else
# ifdef HAVE_SYS_WAIT_H
#  include <sys/wait.h>
# endif
#endif

#include <mystring.h>
#include <logging.h>
#include "bftpdutmp.h"
#include <options.h>
#include <login.h>

int global_argc;
char **global_argv;
struct sockaddr_in name;
int isparent = 1;
int listensocket, sock;
FILE *passwdfile = NULL, *groupfile = NULL, *devnull;
struct sockaddr_in remotename;
char *remotehostname;
char noatexit = 0;
int control_timeout, data_timeout;
int alarm_type = 0;

/* Command line parameters */
char *configpath = PATH_BFTPD_CONF;
char daemonmode = 0;

void print_file(int number, char *filename)
{
	FILE *phile;
	char foo[256];
	phile = fopen(filename, "r");
	if (phile) {
		while (fgets(foo, sizeof(foo), phile)) {
			foo[strlen(foo) - 1] = '\0';
			control_printf(SL_SUCCESS, "%i-%s", number, foo);
		}
		fclose(phile);
	}
}

void end_child()
{
    if (noatexit)
        return;
	if (passwdfile)
		fclose(passwdfile);
	if (groupfile)
		fclose(groupfile);
	config_end();
	bftpd_log("Quitting.\n");
    bftpd_statuslog(1, 0, "quit");
    bftpdutmp_end();
	log_end();
	login_end();
	if (daemonmode)
		close(sock);
}

void end_parent()
{								/* Called in standalone mode before termination of the */
	if (!isparent)				/* parent process. */
		return;
	close(listensocket);
}

void handler_sigchld(int sig)
{
	wait(NULL);					/* Get the child's return code so that the zombie dies */
}

void handler_sigterm(int signum)
{
	exit(0);					/* Force normal termination so that end_child() is called */
}

void handler_sigalrm(int signum)
{
    if (alarm_type) {
        close(alarm_type);
        bftpd_log("Kicked from the server due to data connection timeout.\n");
        control_printf(SL_FAILURE, "421 Kicked from the server due to data connection timeout.");
        exit(0);
    } else {
        bftpd_log("Kicked from the server due to control connection timeout.\n");
        control_printf(SL_FAILURE, "421 Kicked from the server due to control connection timeout.");
        exit(0);
    }
}

void init_everything()
{
	if (!daemonmode)
		config_init();
	log_init();
    bftpdutmp_init();
	login_init();
}

int main(int argc, char **argv)
{
	char str[MAXCMD + 1];
	static struct hostent *he;
	int i = 1, port;
	char retval;
	while (((retval = getopt(argc, argv, "c:d"))) > -1) {
		switch (retval) {
			case 'd':
				daemonmode = 1;
				break;
			case 'c':
				configpath = strdup(optarg);
				break;
		}
	}
	if (daemonmode) {
		if (!fork()) {			/* Parent process */
			struct sockaddr_in myaddr, new;
			setsid();
			if (fork())
				return 0;
			signal(SIGCHLD, handler_sigchld);
			chdir("/");
			config_init();
			listensocket = socket(AF_INET, SOCK_STREAM, 0);
#ifdef SO_REUSEADDR
			setsockopt(listensocket, SOL_SOCKET, SO_REUSEADDR, (void *) &i,
					   sizeof(i));
#endif
#ifdef SO_REUSEPORT
			setsockopt(listensocket, SOL_SOCKET, SO_REUSEPORT, (void *) &i,
					   sizeof(i));
#endif
			memset((void *) &myaddr, 0, sizeof(myaddr));
            if (!sscanf(config_getoption("PORT"), "%i", &port))
                port = 21;
			myaddr.sin_port = htons(port);
			if (!strcasecmp(config_getoption("BIND_TO_ADDR"), "any"))
				myaddr.sin_addr.s_addr = INADDR_ANY;
			else
				myaddr.sin_addr.s_addr =
					inet_addr(config_getoption("BIND_TO_ADDR"));
			if (bind(listensocket, (struct sockaddr *) &myaddr, sizeof(myaddr))
				< 0) {
				fprintf(stderr, "Bind failed: %s\n", strerror(errno));
				exit(1);
			}
			if (listen(listensocket, 5)) {
				fprintf(stderr, "Listen failed: %s\n", strerror(errno));
				exit(1);
			}
			for (i = 0; i < 3; i++) {
				close(i);		/* Remove fd pointing to the console */
				open("/dev/null", O_RDWR);	/* Create fd pointing nowhere */
			}
			i = sizeof(new);
			while ((sock = accept(listensocket, (struct sockaddr *) &new, &i))) {
				if (!fork()) {
					close(0);
					close(1);
					close(2);
					isparent = 0;
					dup2(sock, fileno(stdin));
					dup2(sock, fileno(stderr));
					break;
				}
			}
		} else
			exit(0);
	}
	devnull = fopen("/dev/null", "w");
	global_argc = argc;
	global_argv = argv;
	init_everything();
	atexit(end_child);
	signal(SIGTERM, handler_sigterm);
    signal(SIGALRM, handler_sigalrm);
    control_timeout = strtoul(config_getoption("CONTROL_TIMEOUT"), NULL, 0);
    if (!control_timeout)
        control_timeout = 300;
    data_timeout = strtoul(config_getoption("DATA_TIMEOUT"), NULL, 0);
    if (!data_timeout)
        data_timeout = 300;
    xfer_bufsize = strtoul(config_getoption("XFER_BUFSIZE"), NULL, 0);
    if (!xfer_bufsize)
        xfer_bufsize = 4096;

	i = sizeof(remotename);
if (getpeername(fileno(stderr), (struct sockaddr *) &remotename, &i)) {
		control_printf(SL_FAILURE, "421-Could not get peer IP address.\r\n421 %s.",
				 strerror(errno));
		return 0;
	}
	i = 1;
	setsockopt(fileno(stdin), SOL_SOCKET, SO_OOBINLINE, (void *) &i,
			   sizeof(i));
	setsockopt(fileno(stdin), SOL_SOCKET, SO_KEEPALIVE, (void *) &i,
			   sizeof(i));
	/* If option is set, determine the client FQDN */
	if (!strcasecmp((char *) config_getoption("RESOLVE_CLIENT_IP"), "yes")) {
		if ((he = gethostbyaddr((char *) &remotename.sin_addr,
								sizeof(struct in_addr), AF_INET)))
			remotehostname = strdup(he->h_name);
		else
			remotehostname = strdup(inet_ntoa(remotename.sin_addr));
	} else
		remotehostname = strdup(inet_ntoa(remotename.sin_addr));
	bftpd_log("Incoming connection from %s.\n", remotehostname);
    bftpd_statuslog(1, 0, "connect %s", remotehostname);
	i = sizeof(name);
	getsockname(fileno(stdin), (struct sockaddr *) &name, &i);
	print_file(220, config_getoption("MOTD_GLOBAL"));
	/* Parse hello message */
	strcpy(str, (char *) config_getoption("HELLO_STRING"));
	replace(str, "%v", VERSION);
	if (strstr(str, "%h")) {
		if (
			(he =
			 gethostbyaddr((char *) &name.sin_addr, sizeof(struct in_addr),
						   AF_INET))) replace(str, "%h", he->h_name);
		else
			replace(str, "%h", (char *) inet_ntoa(name.sin_addr));
	}
	replace(str, "%i", (char *) inet_ntoa(name.sin_addr));
	control_printf(SL_SUCCESS, "220 %s", str);
	/* Read lines from client and execute appropiate commands */
	while (fgets(str, sizeof(str), stdin)) {
        alarm(control_timeout);
        str[strlen(str) - 2] = 0;
        bftpd_statuslog(2, 0, "%s", str);
#ifdef DEBUG
		bftpd_log("Processing command: %s\n", str);
#endif
		parsecmd(str);
	}
	return 0;
}
